import pytest
import jwt
import kisee.kisee as kisee
from datetime import datetime, timedelta
import mocks


@pytest.fixture
def settings():
    return kisee.load_conf("tests/test_settings.toml")


@pytest.fixture
def valid_jwt(settings):
    return jwt.encode(
        {
            "iss": "Pouete",
            "sub": "toto",
            "exp": datetime.utcnow() + timedelta(hours=1),
            "jti": "42",
        },
        settings["jwt"]["private_key"],
        algorithm="ES256",
    ).decode("utf8")


@pytest.fixture
def valid_jwt_to_change_pwd(settings):
    return jwt.encode(
        {
            "iss": "Pouete",
            "can_change_pwd": True,
            "sub": "toto",
            "exp": datetime.utcnow() + timedelta(hours=1),
            "jti": "42",
        },
        settings["jwt"]["private_key"],
        algorithm="ES256",
    ).decode("utf8")


@pytest.fixture
def client(loop, aiohttp_client):
    settings = kisee.load_conf("tests/test_settings.toml")
    app = kisee.identification_app(settings)
    return loop.run_until_complete(aiohttp_client(app))


async def test_identification_app(client):
    response = await client.get("/")
    assert response.status == 200


async def test_post_jwt(client):
    response = await client.post(
        "/jwt/", json={"login": "test@localhost.com", "password": "tamtam"}
    )
    assert response.status == 201
    response = await client.post("/jwt/", json={"login": "test@localhost.com"})
    assert response.status == 422


async def test_post_jwt_bad_json(client):
    response = await client.post(
        "/jwt/", data=b"{", headers={"Content-Type": "application/json"}
    )
    assert response.status == 400


async def test_post_jwt_bad_password(client):
    """The test backend consider passwords of less than 4 chars to be wrong.
    """
    response = await client.post(
        "/jwt/", json={"login": "test@localhost.com", "password": ""}
    )
    assert response.status == 403


async def test_get_jwts(client):
    response = await client.get("/jwt/")
    assert response.status == 200


async def test_get_jwt(client):
    response = await client.get("/jwt/xxx")
    assert response.status == 500


async def test_get_users(client):
    response = await client.get("/users/")
    assert response.status == 200


async def test_post_users(client):
    response = await client.post(
        "/users/",
        json={"username": "user", "email": "lol@lol.com", "password": "passwod"},
    )
    assert response.status == 201


async def test_post_users__conflict__user_already_exists(client, monkeypatch):
    monkeypatch.setattr(
        "kisee.providers.demo.DemoBackend.register_user", mocks.register_user
    )
    response = await client.post(
        "/users/",
        json={"username": "user", "password": "password", "email": "test@example.com"},
    )
    assert response.status == 409


async def test_get_forgotten_password(client):
    response = await client.get("/forgotten_passwords/")
    assert response.status == 200


async def test_post_forgotten_password(client, monkeypatch):
    monkeypatch.setattr("kisee.views.send_mail", mocks.send_mail)
    response = await client.post(
        "/forgotten_passwords/", json={"email": "foo@example.com"}
    )
    assert response.status == 201


async def test_post_forgotten_password_empty(client, monkeypatch):
    response = await client.post("/forgotten_passwords/", json={})
    assert response.status == 400


async def test_post_forgotten_password_by_username(client, monkeypatch):
    monkeypatch.setattr("kisee.views.send_mail", mocks.send_mail)
    response = await client.post("/forgotten_passwords/", json={"login": "foo"})
    assert response.status == 201


async def post_forgotten_passwords__bad_request(client):
    """Bad request because missing either 'email' or 'username' field
    """
    response = await client.post(
        "/forgotten_passwords/", json={"some-useless-field": "foo"}
    )
    assert response.status == 400


async def test_post_users__bad_request__missing_required_fields(client):
    response = await client.post("/users/", json={"username": "only-username"})
    assert response.status == 400


async def test_post_users__bad_request__invalid_email(client):
    response = await client.post(
        "/users/", json={"username": "user", "email": "lol", "password": "passwod"}
    )
    assert response.status == 400


async def test_patch_users(client):
    response = await client.patch(
        "/users/test/",
        headers={"Authorization": "Basic dGVzdDp0ZXN0"},
        json={"password": "passwod"},
    )
    assert response.status == 204


async def test_patch_users_with_jwt(client, valid_jwt):
    """A normal JWT is not enough to change a password: it does not proove
    the user knows its password.

    To change a password, one must either have:
    - A JWT with a can_change_pwd claim (from "I forgot my password").
    - Auth using basic auth to proove it knows its password.
    """
    response = await client.patch(
        "/users/toto/",
        headers={"Authorization": "Bearer " + valid_jwt},
        json={"password": "passwod"},
    )
    assert response.status == 403


async def test_patch_wrong_user_with_jwt(client, valid_jwt_to_change_pwd):
    response = await client.patch(
        "/users/admin/",
        headers={"Authorization": "Bearer " + valid_jwt_to_change_pwd},
        json={"password": "passwod"},
    )
    assert response.status == 403


async def test_patch_users_missing_auth(client):
    response = await client.patch("/users/test/", json={"password": "passwod"})
    assert response.status == 401


async def test_patch_users_bad_auth(client):
    response = await client.patch(
        "/users/test/",
        headers={"Authorization": "I AM ROOT TAHHTS IT"},
        json={"password": "passwod"},
    )
    assert response.status == 401


async def test_patch_users__bad_request__missing_field(client):
    """Missing 'password' field in json input
    """
    response = await client.patch(
        "/users/test/",
        headers={"Authorization": "Basic dGVzdDp0ZXN0"},
        json={"some-useless-field": "foo"},
    )
    assert response.status == 400


async def test_health(client):
    response = await client.get("/health/")
    assert response.status == 200



"""Some utils for kisee
"""
from kisee.identity_provider import IdentityProvider, User


async def get_user_with_email_or_username(
    user_input: dict, idp_backend: IdentityProvider
) -> User:
    """Retrieve user with either email or username
    """
    if "email" in user_input:
        return await idp_backend.get_user_by_email(user_input["email"])
    return await idp_backend.get_user_by_username(user_input["login"])



"""Views for the IdP server, implementing:

- GET /
- GET /jwt/
- POST /jwt/
"""

import json
import logging
from datetime import datetime, timedelta

import coreapi
import jwt
import psutil
import shortuuid
from aiohttp import web

from kisee.authentication import authenticate_user
from kisee.emails import forge_forgotten_email, is_email, send_mail
from kisee.identity_provider import UserAlreadyExist
from kisee.serializers import serialize
from kisee.utils import get_user_with_email_or_username

logger = logging.getLogger(__name__)


async def get_root(request: web.Request) -> web.Response:
    """https://tools.ietf.org/html/draft-nottingham-json-home-06
    """
    hostname = request.app["settings"]["server"]["hostname"]
    home = {
        "api": {
            "title": "Identification Provider",
            "links": {
                "author": "mailto:julien@palard.fr",
                "describedBy": "https://kisee.readthedocs.io",
            },
        },
        "resources": {
            "jwt": {
                "href": f"{hostname}/jwt/",
                "hints": {
                    "allow": ["GET", "POST"],
                    "formats": {"application/coreapi+json": {}},
                },
            },
            "users": {
                "href": f"{hostname}/users/",
                "hints": {
                    "allow": ["GET", "POST", "PATCH"],
                    "formats": {"application/coreapi+json": {}},
                },
            },
            "forgotten-passwords": {
                "href": f"{hostname}/forgotten-passwords/",
                "hints": {
                    "allow": ["GET", "POST"],
                    "formats": {"application/coreapi+json": {}},
                },
            },
        },
        "actions": {
            "register-user": {
                "href": f"{hostname}/users/",
                "method": "POST",
                "fields": [
                    {"name": "username", "required": True},
                    {"name": "password", "required": True},
                    {"name": "email", "required": True},
                ],
            },
            "create-token": {
                "href": f"{hostname}/jwt/",
                "method": "POST",
                "fields": [
                    {"name": "login", "required": True},
                    {"name": "password", "required": True},
                ],
            },
        },
    }
    return web.Response(
        body=json.dumps(home, indent=4), content_type="application/json-home"
    )


async def get_users(request: web.Request) -> web.Response:
    """View for GET /users/, just describes that a POST is possible.
    """
    hostname = request.app["settings"]["server"]["hostname"]
    return serialize(
        request,
        coreapi.Document(
            url=f"{hostname}/users/",
            title="Users",
            content={
                "users": [],
                "register_user": coreapi.Link(
                    action="post",
                    title="Register a new user",
                    description="POSTing to this endpoint creates a new user",
                    fields=[
                        coreapi.Field(name="username", required=True),
                        coreapi.Field(name="password", required=True),
                        coreapi.Field(name="email", required=True),
                    ],
                ),
            },
        ),
    )


async def post_users(request: web.Request) -> web.Response:
    """A client is asking to create a new user
    """
    data = await request.json()

    if not all(key in data.keys() for key in {"username", "email", "password"}):
        raise web.HTTPBadRequest(reason="Missing required input fields")

    logger.debug("Trying to create user %s", data["username"])

    if not is_email(data["email"]):
        raise web.HTTPBadRequest(reason="Email is not valid")

    try:
        await request.app["identity_backend"].register_user(
            data["username"], data["password"], data["email"]
        )
    except UserAlreadyExist:
        raise web.HTTPConflict(reason="User already exist")

    location = f"/users/{data['username']}/"
    return web.Response(status=201, headers={"Location": location})


async def patch_user(request: web.Request) -> web.Response:
    """Patch user password
    """
    user, claims = await authenticate_user(request)
    if not claims.get("can_change_pwd"):
        raise web.HTTPForbidden(reason="Password change forbidden")
    data = await request.json()
    if "password" not in data:
        raise web.HTTPBadRequest(reason="Missing fields to patch")
    username = request.match_info["username"]
    if username != user.username:
        raise web.HTTPForbidden(reason="Token does not apply to user resource")
    await request.app["identity_backend"].set_password_for_user(user, data["password"])
    return web.Response(status=204)


async def get_jwts(request: web.Request) -> web.Response:
    """Handlers for GET /jwt/, just describes that a POST is possible.
    """
    hostname = request.app["settings"]["server"]["hostname"]
    return serialize(
        request,
        coreapi.Document(
            url=f"{hostname}/jwt/",
            title="JSON Web Tokens",
            content={
                "tokens": [],
                "add_token": coreapi.Link(
                    action="post",
                    title="Create a new JWT",
                    description="POSTing to this endpoint create JWT tokens.",
                    fields=[
                        coreapi.Field(name="login", required=True),
                        coreapi.Field(name="password", required=True),
                    ],
                ),
            },
        ),
    )


async def get_jwt(request: web.Request) -> web.Response:
    """Handler for GET /jwt{/jid}.
    """
    del request  # unused
    raise NotImplementedError()


async def post_jwt(request: web.Request) -> web.Response:
    """A user is asking for a JWT.
    """
    data = await request.json()
    if "login" not in data or "password" not in data:
        raise web.HTTPUnprocessableEntity(reason="Missing login or password.")
    logger.debug("Trying to identify user %s", data["login"])
    user = await request.app["identity_backend"].identify(
        data["login"], data["password"]
    )
    if user is None:
        raise web.HTTPForbidden(reason="Failed identification for kisee.")
    jti = shortuuid.uuid()
    return serialize(
        request,
        coreapi.Document(
            url="/jwt/",
            title="JSON Web Tokens",
            content={
                "tokens": [
                    jwt.encode(
                        {
                            "iss": request.app["settings"]["jwt"]["iss"],
                            "sub": user.user_id,
                            "exp": datetime.utcnow() + timedelta(hours=1),
                            "jti": jti,
                        },
                        request.app["settings"]["jwt"]["private_key"],
                        algorithm="ES256",
                    ).decode("utf8")
                ],
                "add_token": coreapi.Link(
                    action="post",
                    title="Create a new JWT",
                    description="POSTing to this endpoint create JWT tokens.",
                    fields=[
                        coreapi.Field(name="login", required=True),
                        coreapi.Field(name="password", required=True),
                    ],
                ),
            },
        ),
        status=201,
        headers={"Location": "/jwt/" + jti},
    )


async def get_forgotten_passwords(request: web.Request) -> web.Response:
    """Get forgotten password view, just describes that a POST is possible.
    """
    hostname = request.app["settings"]["server"]["hostname"]
    return serialize(
        request,
        coreapi.Document(
            url=f"{hostname}/forgotten-passwords/",
            title="Forgotten password management",
            content={
                "reset-password": coreapi.Link(
                    action="post",
                    title="",
                    description="""
                        POSTing to this endpoint subscribe for a forgotten password
                    """,
                    fields=[coreapi.Field(name="login"), coreapi.Field(name="email")],
                )
            },
        ),
    )


async def post_forgotten_passwords(request: web.Request) -> web.Response:
    """Create process to register new password
    """
    data = await request.json()
    if "login" not in data and "email" not in data:
        raise web.HTTPBadRequest(reason="Missing required fields email or login")

    user = await get_user_with_email_or_username(data, request.app["identity_backend"])
    jwt_token = jwt.encode(
        {
            "iss": request.app["settings"]["jwt"]["iss"],
            "sub": user.username,
            "exp": datetime.utcnow() + timedelta(hours=12),
            "jti": shortuuid.uuid(),
            "can_change_pwd": True,
        },
        request.app["settings"]["jwt"]["private_key"],
        algorithm="ES256",
    ).decode("utf-8")
    content_text, content_html = forge_forgotten_email(
        user.username, user.email, jwt_token
    )
    subject = "Forgotten password"
    send_mail(
        subject,
        content_text,
        content_html,
        request.app["settings"]["email"],
        user.email,
    )
    return web.Response(status=201)


async def get_health(request: web.Request) -> web.Response:
    """Get service health metrics
    """
    is_database_ok = (
        "OK" if await request.app["identity_backend"].is_connection_alive() else "KO"
    )
    disk_usage = psutil.disk_usage("/")
    disk_free_percentage = disk_usage.free / disk_usage.total * 100
    return web.Response(
        body=json.dumps(
            {
                "overall": "OK",
                "load_average": open("/proc/loadavg").readline().split(" ")[0],
                "database": is_database_ok,
                "disk_free": f"{disk_free_percentage:.2f}%",
            }
        ),
        content_type="application/json",
    )



"""Tests for the app
"""


import sys
from unittest.mock import patch

import pytest

import kisee.kisee as kisee


def test_load_conf():
    """Test the configuration loading
    """
    with pytest.raises(SystemExit):
        kisee.load_conf("settings.toml")
    config = kisee.load_conf("tests/test_settings.toml")
    assert config["server"]["host"] == "0.0.0.0"


def test_parse_args():
    """Test the argument parsing
    """
    with patch.object(sys, "argv", ["dummy"]):
        args = kisee.parse_args()
        assert args.settings == "settings.toml"
    args = kisee.parse_args(["--settings", "hi.toml"])
    assert args.settings == "hi.toml"



"""Tests for the dumb identity provider
"""


from kisee.providers.demo import DemoBackend


async def test_dumb_idp():
    """Test instanciation and identification
    """
    async with DemoBackend({}) as test_backend:
        identity = await test_backend.identify("dummy_login", "")
        assert identity is None
        identity = await test_backend.identify("dummy_login", "dummy_password")
        assert identity.username == "dummy_login"
        assert (await test_backend.get_user_by_username("toto")).username == "toto"



"""Middlewares for Kisee service
"""

import json
from typing import Callable, Awaitable

from aiohttp import web

from kisee.serializers import serialize


Handler = Callable[[web.Request], Awaitable[web.StreamResponse]]


@web.middleware
async def verify_input_body_is_json(
    request: web.Request, handler: Handler
) -> web.StreamResponse:
    """
    Middleware to verify that input body is of json format
    """
    if request.can_read_body:
        try:
            await request.json()
        except json.decoder.JSONDecodeError:
            raise web.HTTPBadRequest(reason="Malformed JSON.")
    return await handler(request)


@web.middleware
async def coreapi_error_middleware(
    request: web.Request, handler: Handler
) -> web.StreamResponse:
    """Implementation of:
    http://www.coreapi.org/specification/transport/#coercing-4xx-and-5xx-responses-to-errors
    """
    try:
        return await handler(request)
    except web.HTTPException as ex:
        return serialize(
            request, {"_type": "error", "_meta": {"title": ex.reason}}, ex.status
        )



"""Tests for the identification provider class
"""


import pytest

import kisee.identity_provider as idp


def test_import_idp():
    """Test importing identification provider
    """
    with pytest.raises(ImportError):
        idp.import_idp("helloworldxddd~~~")
    with pytest.raises(ImportError):
        idp.import_idp("my.dummy.path")
    with pytest.raises(ImportError):
        idp.import_idp("kisee.kisee.kisee.UnknownClass")
    with pytest.raises(ImportError):
        idp.import_idp("kisee.providers.demo.UnknownClass")
